
/*
    ZyFixer:

    A plug-in to speedup the Zynamics IDA BinDiff plug-in by ~333% by filtering
    out kernel32.dll->FlushFileBuffers() calls.

    Author: Sirmabus 2013
    http://www.sirmabus.macromonkey.com/

    http://www.zynamics.com/bindiff.html

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
    WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
    INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
    (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
    LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    LICENSEE AGREES THAT ANY PORTION OF THIS AGREEMENT FOUND TO BE INVALID OR
    UNENFORCEABLE SHALL BE MODIFIED, TO THE EXTENT ALLOWED BY LAW, SO AS TO ALLOW
    FOR THE ENFORCEMENT OF THE ORIGINAL INTENDED MEANING OF THE PORTION FOUND TO BE
    INVALID OR UNENFORCEABLE.

    Copyright and license acknowledgments:
    BinDiff & Zynamics copyrighted and, or, trademarked names et al
    of zynamics/Google Inc.
*/

#define WIN32_LEAN_AND_MEAN
#define WINVER       0x0502
#define _WIN32_WINNT 0x0502
#define _WIN32_WINDOWS 0x0502
#define _WIN32_IE 0x0601
#include <windows.h>
#include <intrin.h>
#include <psapi.h>
#pragma comment(lib, "psapi.lib")

// IDA SDK
#include <ida.hpp>
#include <idp.hpp>
#include <loader.hpp>
#include <diskio.hpp>
#pragma comment(lib, "ida.lib")

// A commercial and IMHO the best (stable and features-wise) API hooking system lib around.
// You could use one of the several alternatives like MS Detours, EasyHook, etc.
namespace MADCODEHOOK
{
    #include <madCHook.h>
    #pragma comment(lib, "madCHook.lib")
};

#define SIZESTR(x) (sizeof(x) - 1)

// ======================================================================================

// Hook kernel32.dll->FlushFileBuffers() to do nothing when called form zynamics(tm) plug-ins
static PVOID pFlushFileBuffersNext = NULL;
static UINT  uZFlushFileBuffersCalls = 0;
static BOOL WINAPI _FlushFileBuffers(__in HANDLE hFile)
{
    // Get caller module by name
    char szCaller[_MAX_FNAME + _MAX_EXT]; szCaller[0] = szCaller[SIZESTR(szCaller)] = 0;
    HMODULE hCallerModule = NULL;
    GetModuleHandleEx((GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT | GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS), (LPCTSTR) _ReturnAddress(), &hCallerModule);
    if(hCallerModule) GetModuleBaseName((HANDLE) -1, hCallerModule, szCaller, SIZESTR(szCaller));
    //msg("_FlushFileBuffers: %08X %08X:%08X %s\n", uZFlushFileBuffersCalls, hCallerModule, _ReturnAddress(), szCaller);

    // Skip if in Zy module space
    if(hCallerModule && (_strnicmp(szCaller, "zynamics_", SIZESTR("zynamics_")) == 0))
    {
        uZFlushFileBuffersCalls++;
        return(TRUE);
    }
    else
        // Else chain original
        return(((BOOL (WINAPI *)(HANDLE)) pFlushFileBuffersNext)(hFile));
}


// ======================================================================================
static int OnInitialize()
{
    MADCODEHOOK::InstallMadCHook();
    msg(">> ZyFixer (%s) By Sirmabus - see: \"IDA_ZyFixer.txt\" for more info <<\n", __DATE__);

    // Apply API hook
    if(!MADCODEHOOK::HookAPI("kernel32.dll", "FlushFileBuffers", _FlushFileBuffers, &pFlushFileBuffersNext, DONT_COUNT))
        msg("** ZyBooster: Failed to hook \"FlushFileBuffers()\"!, load aborted **\n");

    return(PLUGIN_KEEP);
}

// ======================================================================================
static void OnTerminate()
{
    // Shouldn't be necessary but unhook anyhow
    if(pFlushFileBuffersNext)
        MADCODEHOOK::UnhookAPI(&pFlushFileBuffersNext);

    MADCODEHOOK::UninstallMadCHook();
}

// ======================================================================================
// When invoked (via the IDA Edit -> Plugins menu) just show total flush calls averted
static void OnRun(int)
{
    msg("Total Z FlushFileBuffers() calls: %u\n", uZFlushFileBuffersCalls);
    uZFlushFileBuffersCalls = 0;
}


// ======================================================================================

// Plug-in description block
extern "C" __declspec(align(16)) plugin_t PLUGIN =
{
    IDP_INTERFACE_VERSION,
    PLUGIN_FIX,
    OnInitialize,
    OnTerminate,
    OnRun,
    "IDA ZyFixer plug-in By Sirmabus",
    "",
    "ZyFixer By Sirmabus",
    ""
};